use std::sync::Arc;

use actix_web::{
    error::InternalError, get, middleware::Logger, post, web, App, HttpResponse, HttpServer,
    Responder,
};
use ledger_utxo::{AccountId, AssetDefinition, AssetManager, Status, TransactionId};
use serde::{Deserialize, Serialize};
use serde_json::json;

#[derive(Deserialize)]
pub struct Movement {
    pub account: AccountId,
    pub amount: String,
    pub asset: String,
}

#[derive(Deserialize)]
pub struct Deposit {
    pub account: AccountId,
    pub amount: String,
    pub asset: String,
    pub memo: String,
    pub status: Status,
}

impl Deposit {
    pub async fn to_ledger_transaction(
        self,
        ledger: &Ledger,
    ) -> Result<ledger_utxo::Transaction, ledger_utxo::Error> {
        let amount = ledger._inner.parse_amount(&self.asset, &self.amount)?;
        ledger
            ._inner
            .deposit(&self.account, amount, self.status, self.memo)
            .await
    }
}

#[derive(Deserialize)]
pub struct Transaction {
    pub debit: Vec<Movement>,
    pub credit: Vec<Movement>,
    pub memo: String,
    pub status: Status,
}

impl Transaction {
    pub async fn to_ledger_transaction(
        self,
        ledger: &Ledger,
    ) -> Result<ledger_utxo::Transaction, ledger_utxo::Error> {
        todo!()
        /*ledger
        ._inner
        .new_transaction(self.memo, self.status, from, to)
        .await*/
    }
}

#[derive(Serialize, Deserialize)]
struct Item {
    id: i32,
    name: String,
}

#[derive(Serialize)]
struct AccountResponse {
    amount: String,
    asset: Arc<str>,
}

#[get("/balance/{id}")]
async fn get_balance(info: web::Path<AccountId>, ledger: web::Data<Ledger>) -> impl Responder {
    let asset_manager = ledger._inner.asset_manager();
    match ledger._inner.get_balance(&info.0).await {
        Ok(balances) => {
            match balances
                .into_iter()
                .map(|amount| {
                    asset_manager
                        .get_name(amount.asset().id)
                        .map(|asset| AccountResponse {
                            amount: amount.to_string(),
                            asset: asset,
                        })
                })
                .collect::<Result<Vec<_>, _>>()
            {
                Ok(balances) => HttpResponse::Ok().json(balances),
                Err(e) => HttpResponse::BadRequest().json(e),
            }
        }
        Err(e) => HttpResponse::BadRequest().json(e),
    }
}

#[get("/tx/{id}")]
async fn get_transaction(
    info: web::Path<TransactionId>,
    ledger: web::Data<Ledger>,
) -> impl Responder {
    // Retrieve an item by ID from a database or another data source.
    // For this example, we'll return a hardcoded item.

    let tx = ledger._inner.get_transaction(&info.0).await;
    println!("{:?}", tx);

    if let Ok(tx) = tx {
        HttpResponse::Ok().json(tx)
    } else {
        HttpResponse::BadRequest().json(json!({"error": "not found"}))
    }
}

#[post("/deposit")]
async fn deposit(item: web::Json<Deposit>, ledger: web::Data<Ledger>) -> impl Responder {
    // Insert the item into a database or another data source.
    // For this example, we'll just echo the received item.
    match item.into_inner().to_ledger_transaction(&ledger).await {
        Ok(tx) => {
            // Insert the item into a database or another data source.
            // For this example, we'll just echo the received item.
            HttpResponse::Created().json(tx)
        }
        Err(e) => HttpResponse::Created().json(e),
    }
}

#[post("/tx")]
async fn create_transaction(
    item: web::Json<Transaction>,
    ledger: web::Data<Ledger>,
) -> impl Responder {
    if let Ok(tx) = item.into_inner().to_ledger_transaction(&ledger).await {
        // Insert the item into a database or another data source.
        // For this example, we'll just echo the received item.
        HttpResponse::Created().json(tx)
    } else {
        HttpResponse::Created().json(json!({"test": "true"}))
    }
}

pub struct Ledger {
    _inner: ledger_utxo::Ledger<'static, ledger_utxo::Batch<'static>, ledger_utxo::Sqlite<'static>>,
}

#[actix_web::main]
async fn main() -> std::io::Result<()> {
    if std::env::var_os("RUST_LOG").is_none() {
        std::env::set_var("RUST_LOG", "actix_web=info");
    }
    env_logger::init_from_env(env_logger::Env::new().default_filter_or("info"));

    let asset_manager = AssetManager::new(vec![
        AssetDefinition::new(1, "BTC", 8),
        AssetDefinition::new(2, "USD", 4),
    ]);

    let pool = ledger_utxo::sqlite::SqlitePoolOptions::new()
        .connect("sqlite:./test.db")
        .await
        .expect("pool");

    let storage = ledger_utxo::Sqlite::new(pool.clone(), asset_manager.clone());
    storage.setup().await.expect("setup");

    HttpServer::new(move || {
        let storage = ledger_utxo::Sqlite::new(pool.clone(), asset_manager.clone());
        let ledger = ledger_utxo::Ledger::new(storage, asset_manager.clone());
        App::new()
            .wrap(Logger::default())
            .app_data(web::Data::new(Ledger { _inner: ledger }))
            .app_data(web::JsonConfig::default().error_handler(|err, _req| {
                InternalError::from_response(
                    "",
                    HttpResponse::BadRequest()
                        .content_type("application/json")
                        .body(format!(r#"{{"error":"{}"}}"#, err)),
                )
                .into()
            }))
            .service(get_balance)
            .service(get_transaction)
            .service(deposit)
            .service(create_transaction)
    })
    .bind("127.0.0.1:8080")?
    .run()
    .await
}
