use crate::{amount::AmountCents, asset::AssetId, Amount, Asset, Error};
use std::{collections::HashMap, sync::Arc};

#[derive(Clone, Debug, Eq, PartialEq)]
pub struct AssetManager {
    assets: Arc<HashMap<AssetId, AssetDefinition>>,
    asset_names: Arc<HashMap<Arc<str>, AssetDefinition>>,
}

#[derive(Clone, Debug, Eq, PartialEq, serde::Deserialize)]
pub struct AssetDefinition {
    #[serde(flatten)]
    pub asset: Asset,
    name: Arc<str>,
}

impl AssetManager {
    pub fn new(assets: Vec<AssetDefinition>) -> Self {
        Self {
            assets: Arc::new(
                assets
                    .iter()
                    .map(|asset| (asset.asset.id, asset.clone()))
                    .collect(),
            ),
            asset_names: Arc::new(
                assets
                    .into_iter()
                    .map(|asset| (asset.name.clone(), asset))
                    .collect(),
            ),
        }
    }

    pub fn get_name(&self, id: AssetId) -> Result<Arc<str>, Error> {
        self.assets
            .get(&id)
            .map(|asset| asset.name.clone())
            .ok_or(Error::AssetIdNotFound(id))
    }

    pub fn asset(&self, id: AssetId) -> Result<Asset, Error> {
        self.assets
            .get(&id)
            .map(|asset| asset.asset)
            .ok_or(Error::AssetIdNotFound(id))
    }

    pub fn human_amount_by_name(&self, name: &str, human_amount: &str) -> Result<Amount, Error> {
        Ok(self
            .asset_names
            .get(name)
            .map(|asset| Amount::from_human(asset.asset, human_amount))
            .ok_or(Error::AssetNotFound(name.to_owned()))??)
    }

    pub fn amount_by_and_cents(&self, id: AssetId, cents: AmountCents) -> Result<Amount, Error> {
        self.asset(id).map(|asset| Amount::new(asset, cents))
    }
}

impl AssetDefinition {
    pub fn new(id: u32, name: &str, precision: u8) -> Self {
        Self {
            asset: Asset { id, precision },
            name: name.into(),
        }
    }
}
