use crate::{
    amount::AmountCents,
    asset::AssetId,
    transaction::{from_db, Type},
    AccountId, Amount, Payment, PaymentId, Status, Transaction, TransactionId,
};

#[derive(thiserror::Error, Debug)]
pub enum Error {
    #[error("Storage error: {0}")]
    Storage(String),

    #[error("Invalid date format: {0}")]
    InvalidDate(String),

    #[error("Spend payment: {0}")]
    SpendPayment(String),

    #[error("No storage update when expecting")]
    NoUpdate,

    #[error("Record not found")]
    NotFound,

    /// TODO: Convert the AmountCents to Amount for better error reporting upstream
    #[error("Not enough unspent payments (missing {0} cents)")]
    NotEnoughUnspentPayments(AmountCents),
}

#[async_trait::async_trait]
pub trait Batch<'a> {
    async fn rollback(self) -> Result<(), Error>;

    async fn commit(self) -> Result<(), Error>;

    async fn spend_payment(
        &mut self,
        payment_id: &PaymentId,
        status: Status,
        transaction_id: &TransactionId,
    ) -> Result<(), Error>;

    async fn get_payment_status(
        &mut self,
        transaction_id: &TransactionId,
    ) -> Result<Option<Status>, Error>;

    async fn store_new_payment(&mut self, payment: &Payment) -> Result<(), Error>;

    async fn store_transaction(&mut self, transaction: &Transaction) -> Result<(), Error>;

    async fn relate_account_to_transaction(
        &mut self,
        transaction: &Transaction,
        account: &AccountId,
    ) -> Result<(), Error>;
}

#[async_trait::async_trait]
pub trait Storage<'a, B>
where
    B: Batch<'a>,
{
    /// Begins a transaction
    ///
    /// A transaction prepares the storage for a batch of operations, where all
    /// the operations are persisted or none.
    ///
    /// The transaction is returned as a Batch object, which is used to perform
    /// the changes in the transactions and payments.
    ///
    /// The batch has also a rollback
    async fn begin(&'a self) -> Result<B, Error>;

    /// Returns a payment object by ID.
    async fn get_payment(&self, id: PaymentId) -> Result<Payment, Error>;

    /// Similar to get_payment() but errors if the requested payment is not spendable.
    async fn get_unspent_payment(&self, id: PaymentId) -> Result<Payment, Error> {
        let payment = self.get_payment(id).await?;
        if payment.is_spendable() {
            Ok(payment)
        } else {
            Err(Error::NotFound)
        }
    }

    /// Returns the balances for a given account
    async fn get_balance(&self, account: &AccountId) -> Result<Vec<Amount>, Error>;

    /// Returns a list of unspent payments for a given account and asset.
    ///
    /// The payments should be returned sorted by ascending amount, this bit is
    /// important to make sure that all negative payments are always taken into
    /// account. It will also improve the database by using many small payments
    /// instead of a few large ones, which will make the database faster leaving
    /// fewer unspent payments when checking for balances.
    async fn get_unspent_payments(
        &self,
        account: &AccountId,
        asset: AssetId,
        target_amount: AmountCents,
    ) -> Result<Vec<Payment>, Error>;

    /// Returns a transaction object by id
    async fn get_transaction(
        &self,
        transaction_id: &TransactionId,
    ) -> Result<from_db::Transaction, Error>;

    /// Returns a list of a transactions for a given account (and optionally
    /// filter by types). The list of transactions is expected to be sorted by
    /// date desc (newest transactions first)
    async fn get_transactions(
        &self,
        account: &AccountId,
        typ: Option<Type>,
    ) -> Result<Vec<from_db::Transaction>, Error>;
}
