use super::{deposit, get_instance};
use crate::{AccountId, Status};

#[tokio::test]
async fn pending_deposit_and_failure() {
    let source = "account1".parse::<AccountId>().expect("account");
    let (assets, ledger) = get_instance().await;
    let id = ledger
        .deposit(
            &source,
            assets.amount_by_and_cents(2, 3000).expect("amount"),
            Status::Processing,
            "Test".to_owned(),
        )
        .await
        .expect("valid tx")
        .id()
        .clone();

    assert!(ledger
        .get_balance(&source)
        .await
        .expect("balance")
        .is_empty());

    ledger
        .change_status(&id, Status::Failed)
        .await
        .expect("valid tx");

    assert!(ledger
        .get_balance(&source)
        .await
        .expect("balance")
        .is_empty());
}

#[tokio::test]
async fn deposit_and_transfer() {
    let source = "account1".parse::<AccountId>().expect("account");
    let dest = "account2".parse::<AccountId>().expect("account");
    let fee = "fee".parse::<AccountId>().expect("account");
    let (assets, ledger) = get_instance().await;

    deposit(
        &ledger,
        &source,
        assets.amount_by_and_cents(2, 1000).expect("amount"),
    )
    .await;
    deposit(
        &ledger,
        &source,
        assets.amount_by_and_cents(2, 2000).expect("amount"),
    )
    .await;

    assert_eq!(
        vec![assets.amount_by_and_cents(2, 3000).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );

    ledger
        .new_transaction(
            "Exchange one".to_owned(),
            Status::Settled,
            vec![(
                source.clone(),
                assets.amount_by_and_cents(2, 1300).expect("amount"),
            )],
            vec![
                (
                    dest.clone(),
                    assets.amount_by_and_cents(2, 1250).expect("amount"),
                ),
                (
                    fee.clone(),
                    assets.amount_by_and_cents(2, 50).expect("amount"),
                ),
            ],
        )
        .await
        .expect("valid tx");

    assert_eq!(
        vec![assets.amount_by_and_cents(2, 1700).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );
    assert_eq!(
        vec![assets.amount_by_and_cents(2, 1250).expect("amount")],
        ledger.get_balance(&dest).await.expect("balance")
    );
    assert_eq!(
        vec![assets.amount_by_and_cents(2, 50).expect("amount")],
        ledger.get_balance(&fee).await.expect("balance")
    );
}

#[tokio::test]
async fn balance_decreases_while_pending_spending_and_confirm() {
    let source = "account1".parse::<AccountId>().expect("account");
    let dest = "account2".parse::<AccountId>().expect("account");
    let fee = "fee".parse::<AccountId>().expect("account");
    let (assets, ledger) = get_instance().await;

    deposit(
        &ledger,
        &source,
        assets.amount_by_and_cents(2, 1000).expect("amount"),
    )
    .await;
    deposit(
        &ledger,
        &source,
        assets.amount_by_and_cents(2, 2000).expect("amount"),
    )
    .await;

    assert_eq!(
        vec![assets.amount_by_and_cents(2, 3000).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );

    let id = ledger
        .new_transaction(
            "Exchange one".to_owned(),
            Status::Pending,
            vec![(
                source.clone(),
                assets.amount_by_and_cents(2, 1300).expect("amount"),
            )],
            vec![
                (
                    dest.clone(),
                    assets.amount_by_and_cents(2, 1250).expect("amount"),
                ),
                (
                    fee.clone(),
                    assets.amount_by_and_cents(2, 50).expect("amount"),
                ),
            ],
        )
        .await
        .expect("valid tx")
        .id()
        .clone();

    assert_eq!(
        vec![assets.amount_by_and_cents(2, 1700).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );
    assert!(ledger.get_balance(&dest).await.expect("balance").is_empty());
    assert!(ledger.get_balance(&fee).await.expect("balance").is_empty());

    ledger
        .change_status(&id, Status::Settled)
        .await
        .expect("valid tx");

    assert_eq!(
        vec![assets.amount_by_and_cents(2, 1700).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );
    assert_eq!(
        vec![assets.amount_by_and_cents(2, 1250).expect("amount")],
        ledger.get_balance(&dest).await.expect("balance")
    );
    assert_eq!(
        vec![assets.amount_by_and_cents(2, 50).expect("amount")],
        ledger.get_balance(&fee).await.expect("balance")
    );
}

#[tokio::test]
async fn balance_decreases_while_pending_spending_and_cancel() {
    let source = "account1".parse::<AccountId>().expect("account");
    let dest = "account2".parse::<AccountId>().expect("account");
    let fee = "fee".parse::<AccountId>().expect("account");
    let (assets, ledger) = get_instance().await;

    deposit(
        &ledger,
        &source,
        assets.amount_by_and_cents(2, 1000).expect("amount"),
    )
    .await;
    deposit(
        &ledger,
        &source,
        assets.amount_by_and_cents(2, 2000).expect("amount"),
    )
    .await;

    assert_eq!(
        vec![assets.amount_by_and_cents(2, 3000).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );

    let id = ledger
        .new_transaction(
            "Exchange one".to_owned(),
            Status::Pending,
            vec![(
                source.clone(),
                assets.amount_by_and_cents(2, 1300).expect("amount"),
            )],
            vec![
                (
                    dest.clone(),
                    assets.amount_by_and_cents(2, 1250).expect("amount"),
                ),
                (
                    fee.clone(),
                    assets.amount_by_and_cents(2, 50).expect("amount"),
                ),
            ],
        )
        .await
        .expect("valid tx")
        .id()
        .clone();

    assert_eq!(
        vec![assets.amount_by_and_cents(2, 1700).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );
    assert!(ledger.get_balance(&dest).await.expect("balance").is_empty());
    assert!(ledger.get_balance(&fee).await.expect("balance").is_empty());

    ledger
        .change_status(&id, Status::Cancelled)
        .await
        .expect("valid tx");

    assert_eq!(
        vec![assets.amount_by_and_cents(2, 3000).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );
    assert!(ledger.get_balance(&dest).await.expect("balance").is_empty());
    assert!(ledger.get_balance(&fee).await.expect("balance").is_empty());
}

#[tokio::test]
async fn balance_decreases_while_pending_spending_and_failed() {
    let source = "account1".parse::<AccountId>().expect("account");
    let dest = "account2".parse::<AccountId>().expect("account");
    let fee = "fee".parse::<AccountId>().expect("account");
    let (assets, ledger) = get_instance().await;

    deposit(
        &ledger,
        &source,
        assets.amount_by_and_cents(2, 1000).expect("amount"),
    )
    .await;
    deposit(
        &ledger,
        &source,
        assets.amount_by_and_cents(2, 2000).expect("amount"),
    )
    .await;

    assert_eq!(
        vec![assets.amount_by_and_cents(2, 3000).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );

    let tx = ledger
        .new_transaction(
            "Exchange one".to_owned(),
            Status::Pending,
            vec![(
                source.clone(),
                assets.amount_by_and_cents(2, 1300).expect("amount"),
            )],
            vec![
                (
                    dest.clone(),
                    assets.amount_by_and_cents(2, 1250).expect("amount"),
                ),
                (
                    fee.clone(),
                    assets.amount_by_and_cents(2, 50).expect("amount"),
                ),
            ],
        )
        .await
        .expect("valid tx")
        .id()
        .clone();

    assert_eq!(
        vec![assets.amount_by_and_cents(2, 1700).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );
    assert!(ledger.get_balance(&dest).await.expect("balance").is_empty());
    assert!(ledger.get_balance(&fee).await.expect("balance").is_empty());

    ledger
        .change_status(&tx, Status::Processing)
        .await
        .expect("valid tx");

    assert_eq!(
        vec![assets.amount_by_and_cents(2, 1700).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );
    assert!(ledger.get_balance(&dest).await.expect("balance").is_empty());
    assert!(ledger.get_balance(&fee).await.expect("balance").is_empty());

    assert_eq!(
        "Transaction: Status transition from Processing to Cancelled is not allowed".to_owned(),
        ledger
            .change_status(&tx, Status::Cancelled)
            .await
            .unwrap_err()
            .to_string()
    );

    assert_eq!(
        vec![assets.amount_by_and_cents(2, 1700).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );
    assert!(ledger.get_balance(&dest).await.expect("balance").is_empty());
    assert!(ledger.get_balance(&fee).await.expect("balance").is_empty());

    ledger
        .change_status(&tx, Status::Failed)
        .await
        .expect("valid");

    assert_eq!(
        vec![assets.amount_by_and_cents(2, 3000).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );
    assert!(ledger.get_balance(&dest).await.expect("balance").is_empty());
    assert!(ledger.get_balance(&fee).await.expect("balance").is_empty());
}
