use super::{deposit, get_instance};
use crate::{AccountId, Status};

#[tokio::test]
async fn negative_deposit_prevent_spending() {
    let source = "account1".parse::<AccountId>().expect("account");
    let dest = "account2".parse::<AccountId>().expect("account");
    let fee = "fee".parse::<AccountId>().expect("account");
    let (assets, ledger) = get_instance().await;

    // Deposit some money
    deposit(
        &ledger,
        &source,
        assets.amount_by_and_cents(2, 5000).expect("amount"),
    )
    .await;
    // Take money of source's account
    deposit(
        &ledger,
        &source,
        assets.amount_by_and_cents(2, -10000).expect("amount"),
    )
    .await;

    assert_eq!(
        vec![assets.amount_by_and_cents(2, -5000).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );

    // Try to spend money
    assert_eq!(
        "Storage: Not enough unspent payments (missing 6000 cents)".to_owned(),
        ledger
            .new_transaction(
                "Exchange one".to_owned(),
                Status::Settled,
                vec![(
                    source.clone(),
                    assets.amount_by_and_cents(2, 1000).expect("amount")
                )],
                vec![
                    (
                        dest.clone(),
                        assets.amount_by_and_cents(2, 950).expect("amount")
                    ),
                    (
                        fee.clone(),
                        assets.amount_by_and_cents(2, 50).expect("amount")
                    ),
                ],
            )
            .await
            .unwrap_err()
            .to_string()
    );
}

#[tokio::test]
async fn negative_deposit_prevent_spending_payback() {
    let source = "account1".parse::<AccountId>().expect("account");
    let dest = "account2".parse::<AccountId>().expect("account");
    let fee = "fee".parse::<AccountId>().expect("account");
    let (assets, ledger) = get_instance().await;

    // Deposit some money
    deposit(
        &ledger,
        &source,
        assets.amount_by_and_cents(2, 5000).expect("amount"),
    )
    .await;
    // Take money of source's account
    deposit(
        &ledger,
        &source,
        assets.amount_by_and_cents(2, -10000).expect("amount"),
    )
    .await;

    assert_eq!(
        vec![assets.amount_by_and_cents(2, -5000).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );

    // Try to spend money
    assert_eq!(
        "Storage: Not enough unspent payments (missing 6000 cents)".to_owned(),
        ledger
            .new_transaction(
                "Exchange one".to_owned(),
                Status::Settled,
                vec![(
                    source.clone(),
                    assets.amount_by_and_cents(2, 1000).expect("amount")
                )],
                vec![
                    (
                        dest.clone(),
                        assets.amount_by_and_cents(2, 950).expect("amount")
                    ),
                    (
                        fee.clone(),
                        assets.amount_by_and_cents(2, 50).expect("amount")
                    ),
                ],
            )
            .await
            .unwrap_err()
            .to_string()
    );

    // Payback the debt
    deposit(
        &ledger,
        &source,
        assets.amount_by_and_cents(2, 15000).expect("amount"),
    )
    .await;

    ledger
        .new_transaction(
            "Exchange one".to_owned(),
            Status::Settled,
            vec![(
                source.clone(),
                assets.amount_by_and_cents(2, 1000).expect("amount"),
            )],
            vec![
                (
                    dest.clone(),
                    assets.amount_by_and_cents(2, 950).expect("amount"),
                ),
                (
                    fee.clone(),
                    assets.amount_by_and_cents(2, 50).expect("amount"),
                ),
            ],
        )
        .await
        .expect("valid tx");

    assert_eq!(
        vec![assets.amount_by_and_cents(2, 950).expect("amount")],
        ledger.get_balance(&dest).await.expect("balance")
    );
    assert_eq!(
        vec![assets.amount_by_and_cents(2, 50).expect("amount")],
        ledger.get_balance(&fee).await.expect("balance")
    );
    assert_eq!(
        vec![assets.amount_by_and_cents(2, 9000).expect("amount")],
        ledger.get_balance(&source).await.expect("balance")
    );
}
