use crate::{
    amount::AmountCents,
    storage::{Batch, Storage},
    transaction::*,
    AccountId, Amount, Asset, Payment, PaymentId, Status, TransactionId,
};
use chrono::{serde::ts_milliseconds, DateTime, Utc};
use serde::Serialize;
use sha2::{Digest, Sha256};
use std::collections::HashMap;

/// Transactions
///
/// Transactions are the core components of the ledger. The transactions are a
/// list of unspent payments that are about to be spend, to create a new set of
/// Payments, that can be spend in the future. This model is heavily inspired in
/// Bitcoin's UTXO model. The terms in this context are payments, spend and
/// create instead of unspent transactions, input and output.
///
/// This simple architecture allows to track accounts pretty efficiently,
/// because all that matters are unspent payments owned by a given account.
/// Every spent payment is stored for historical reasons but it is not relevant
/// for any calculations regarding available funds.
///
/// The transaction has a few rules, for instance the sum of spend Payments
/// should be the same as create Payments, for each easy. There is no 'fee'
/// concept, so any mismatch in any direction will error the constructor.
///
/// Transactions are immutable after they are finalized, and the payments can
/// only be re-usable if the transaction failed or was cancelled. Once the
/// transaction settles the spent payments are forever spent. Any rollback
/// should be a new transaction, initiated by a higher layer.
///
/// The spent payments are unavailable until the transaction is finalized,
/// either as settled, cancelled or failed. A higher layer should split any
/// available payment to be spend into a new transaction, and then finalize the
/// transaction, and reserve only the exact amount to be spent, otherwise
/// unrelated funds will be held unspentable until the transaction is finalized.
#[derive(Debug, Clone, Serialize)]
pub struct Transaction {
    id: TransactionId,
    #[serde(skip_serializing_if = "Vec::is_empty")]
    spends: Vec<Payment>,
    #[allow(dead_code)]
    reference: String,
    #[serde(rename = "type")]
    typ: Type,
    #[serde(skip_serializing_if = "Vec::is_empty")]
    creates: Vec<Payment>,
    status: Status,
    #[serde(with = "ts_milliseconds")]
    created_at: DateTime<Utc>,
    #[serde(with = "ts_milliseconds")]
    updated_at: DateTime<Utc>,
}

impl Transaction {
    pub fn new_external_withdrawal(
        reference: String,
        status: Status,
        spend: Vec<Payment>,
    ) -> Result<Transaction, Error> {
        let created_at = Utc::now();
        let id = Self::calculate_hash(
            spend.iter().map(|t| &t.id).collect::<Vec<&PaymentId>>(),
            vec![],
            created_at,
        )?;
        let spend = spend
            .into_iter()
            .map(|mut input| {
                input.spent_by = Some(id.clone());
                input
            })
            .collect();
        Ok(Self {
            id,
            spends: spend,
            creates: vec![],
            typ: Type::Withdrawal,
            reference,
            status,
            created_at,
            updated_at: Utc::now(),
        })
    }

    pub fn new_external_deposit(
        reference: String,
        status: Status,
        pay_to: Vec<(AccountId, Amount)>,
    ) -> Result<Transaction, Error> {
        let created_at = Utc::now();
        let id = Self::calculate_hash(
            vec![],
            pay_to
                .iter()
                .map(|t| (&t.0, &t.1))
                .collect::<Vec<(&AccountId, &Amount)>>(),
            created_at,
        )?;
        let create = pay_to
            .into_iter()
            .enumerate()
            .map(|(position, (to, amount))| Payment {
                id: crate::PaymentId {
                    transaction: id.clone(),
                    position,
                },
                to,
                amount,
                spent_by: None,
                status: status.clone(),
            })
            .collect();

        Ok(Self {
            id,
            spends: vec![],
            creates: create,
            reference,
            typ: Type::Deposit,
            status,
            created_at,
            updated_at: Utc::now(),
        })
    }

    pub async fn new(
        reference: String,
        status: Status,
        typ: Type,
        spend: Vec<Payment>,
        pay_to: Vec<(AccountId, Amount)>,
    ) -> Result<Transaction, Error> {
        let created_at = Utc::now();
        let id = Self::calculate_hash(
            spend.iter().map(|t| &t.id).collect::<Vec<&PaymentId>>(),
            pay_to
                .iter()
                .map(|t| (&t.0, &t.1))
                .collect::<Vec<(&AccountId, &Amount)>>(),
            created_at,
        )?;

        for (i, input) in spend.iter().enumerate() {
            if input.spent_by.as_ref() != Some(&id) && !input.is_spendable() {
                return Err(Error::InvalidPaymentStatus(i, input.status.clone()));
            }
        }
        let spend = spend
            .into_iter()
            .map(|mut input| {
                input.spent_by = Some(id.clone());
                input
            })
            .collect();

        let create = pay_to
            .into_iter()
            .enumerate()
            .map(|(position, (to, amount))| Payment {
                id: crate::PaymentId {
                    transaction: id.clone(),
                    position,
                },
                to,
                amount,
                spent_by: None,
                status: status.clone(),
            })
            .collect();

        Ok(Self {
            id,
            reference,
            spends: spend,
            typ,
            creates: create,
            status,
            created_at,
            updated_at: Utc::now(),
        })
    }

    fn calculate_hash(
        spend: Vec<&PaymentId>,
        create: Vec<(&AccountId, &Amount)>,
        created_at: DateTime<Utc>,
    ) -> Result<TransactionId, Error> {
        let mut hasher = Sha256::new();
        let mut spend = spend;

        spend.sort();

        for id in spend.into_iter() {
            hasher.update(&bincode::serialize(id)?);
        }
        for (account, amount) in create.into_iter() {
            hasher.update(&bincode::serialize(account)?);
            hasher.update(&bincode::serialize(amount)?);
        }
        hasher.update(&created_at.timestamp_millis().to_le_bytes());
        Ok(TransactionId::new(hasher.finalize().into()))
    }

    pub async fn settle<'a, B, S>(&mut self, storage: &'a S) -> Result<(), Error>
    where
        B: Batch<'a>,
        S: Storage<'a, B> + Sync + Send,
    {
        self.change_status(Status::Settled)?;
        self.persist::<B, S>(storage).await
    }

    #[inline]
    pub fn change_status(&mut self, new_status: Status) -> Result<(), Error> {
        if self.status.can_transition_to(&new_status) {
            self.spends.iter_mut().for_each(|payment| {
                payment.status = new_status.clone();
                if new_status.is_rollback() {
                    payment.spent_by = None;
                }
            });
            self.creates.iter_mut().for_each(|payment| {
                payment.status = new_status.clone();
            });
            self.status = new_status;
            Ok(())
        } else {
            Err(Error::StatusTransitionNotAllowed(
                self.status.clone(),
                new_status,
            ))
        }
    }

    #[inline]
    fn check_no_negative_amounts_are_spent(
        &self,
        debit: &HashMap<Asset, i128>,
    ) -> Result<(), Error> {
        for (asset, amount) in debit.iter() {
            if *amount <= 0 {
                return Err(Error::InvalidAmount(
                    asset.new_amount(*amount),
                    asset.new_amount(*amount),
                ));
            }
        }

        Ok(())
    }

    pub(crate) fn validate(&self) -> Result<(), Error> {
        let calculated_id = Self::calculate_hash(
            self.spends.iter().map(|p| &p.id).collect::<Vec<_>>(),
            self.creates
                .iter()
                .map(|p| (&p.to, &p.amount))
                .collect::<Vec<_>>(),
            self.created_at,
        )?;

        if calculated_id != self.id {
            return Err(Error::InvalidTransactionId(self.id.clone(), calculated_id));
        }

        let mut debit = HashMap::<Asset, AmountCents>::new();
        let mut credit = HashMap::<Asset, AmountCents>::new();

        for (i, input) in self.spends.iter().enumerate() {
            if input.spent_by.is_some() && input.spent_by.as_ref() != Some(&self.id) {
                return Err(Error::SpentPayment(i));
            }
            if let Some(value) = debit.get_mut(input.amount.asset()) {
                *value = input
                    .amount
                    .cents()
                    .checked_add(*value)
                    .ok_or(Error::Overflow)?;
            } else {
                debit.insert(*input.amount.asset(), input.amount.cents());
            }
        }

        self.check_no_negative_amounts_are_spent(&debit)?;

        if !self.typ.is_transaction() {
            // We don't care input/output balance in external operations
            // (withdrawals/deposits), because these operations are inbalanced
            return Ok(());
        }

        for (i, output) in self.creates.iter().enumerate() {
            if output.spent_by.is_some() {
                return Err(Error::SpentPayment(i));
            }
            if let Some(value) = credit.get_mut(output.amount.asset()) {
                *value = output
                    .amount
                    .cents()
                    .checked_add(*value)
                    .ok_or(Error::Overflow)?;
            } else {
                credit.insert(*output.amount.asset(), output.amount.cents());
            }
        }

        for (asset, credit_amount) in credit.into_iter() {
            if let Some(debit_amount) = debit.remove(&asset) {
                if debit_amount != credit_amount {
                    return Err(Error::InvalidAmount(
                        asset.new_amount(debit_amount),
                        asset.new_amount(credit_amount),
                    ));
                }
            } else {
                return Err(Error::MissingSpendingAsset(asset));
            }
        }

        if let Some((asset, _)) = debit.into_iter().next() {
            return Err(Error::MissingPaymentAsset(asset));
        }

        Ok(())
    }

    pub fn spends(&self) -> &[Payment] {
        &self.spends
    }

    pub fn creates(&self) -> &[Payment] {
        &self.creates
    }

    pub fn id(&self) -> &TransactionId {
        &self.id
    }

    pub fn status(&self) -> &Status {
        &self.status
    }

    pub fn typ(&self) -> &Type {
        &self.typ
    }

    pub fn reference(&self) -> &str {
        &self.reference
    }

    pub fn created_at(&self) -> DateTime<Utc> {
        self.created_at
    }

    pub fn updated_at(&self) -> DateTime<Utc> {
        self.updated_at
    }

    pub async fn persist<'a, B, S>(&mut self, storage: &'a S) -> Result<(), Error>
    where
        B: Batch<'a>,
        S: Storage<'a, B> + Sync + Send,
    {
        let mut batch = storage.begin().await?;
        if let Some(status) = batch.get_payment_status(&self.id).await? {
            if status.is_finalized() {
                return Err(Error::TransactionUpdatesNotAllowed);
            }
        }
        self.validate()?;
        self.updated_at = Utc::now();
        batch.store_transaction(self).await?;
        for payment in self.creates.iter() {
            batch.store_new_payment(payment).await?;
            batch
                .relate_account_to_transaction(&self, &payment.to)
                .await?;
        }
        for input in self.spends.iter() {
            batch
                .spend_payment(&input.id, self.status.clone(), &self.id)
                .await?;
            batch
                .relate_account_to_transaction(&self, &input.to)
                .await?;
        }
        batch.commit().await?;
        Ok(())
    }
}

impl TryFrom<from_db::Transaction> for Transaction {
    type Error = Error;

    fn try_from(value: from_db::Transaction) -> Result<Self, Self::Error> {
        let tx = Transaction {
            id: value.id,
            typ: value.typ,
            spends: value.spend,
            creates: value.create,
            reference: value.reference,
            status: value.status,
            created_at: value.created_at,
            updated_at: value.updated_at,
        };
        tx.validate()?;
        Ok(tx)
    }
}
